/*
 * Copyright (C) 2017, David PHAM-VAN <dev.nfet.net@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// ignore_for_file: omit_local_variable_types

import 'dart:io';

import 'package:pdfio/pdfio.dart';

void printText(PdfGraphics canvas, String text, PFont font, double top) {
  text = text + font.fontName;
  const double fontSize = 20;
  final PdfFontMetrics metrics = font.stringMetrics(text) * fontSize;

  const double deb = 5;

  const double x = 50;
  final double y = canvas.page.pageFormat.height - top;

  canvas
    ..drawRect(x + metrics.left, y + metrics.top, metrics.width, metrics.height)
    ..setColor(const PdfColor(0.9, 0.9, 0.9))
    ..fillPath()
    ..drawLine(x + metrics.left - deb, y, x + metrics.right + deb, y)
    ..setColor(PdfColors.blue)
    ..strokePath()
    ..drawLine(x + metrics.left - deb, y + metrics.ascent,
        x + metrics.right + deb, y + metrics.ascent)
    ..setColor(PdfColors.green)
    ..strokePath()
    ..drawLine(x + metrics.left - deb, y + metrics.descent,
        x + metrics.right + deb, y + metrics.descent)
    ..setColor(PdfColors.purple)
    ..strokePath()
    ..setColor(const PdfColor(0.3, 0.3, 0.3))
    ..drawString(font, fontSize, text, x, y);
}

void main() {
  final PDocument pdf = PDocument();
  final PPage page =
  PPage(pdf, pageFormat: const PPageFormat(500, 430));

  final PdfGraphics g = page.getGraphics();
  int top = 0;
  const String s = 'Hello ';

  printText(g, s, PFont.courier(pdf), 20.0 + 30.0 * top++);
  printText(g, s, PFont.courierBold(pdf), 20.0 + 30.0 * top++);
  printText(g, s, PFont.courierOblique(pdf), 20.0 + 30.0 * top++);
  printText(g, s, PFont.courierBoldOblique(pdf), 20.0 + 30.0 * top++);

  printText(g, s, PFont.helvetica(pdf), 20.0 + 30.0 * top++);
  printText(g, s, PFont.helveticaBold(pdf), 20.0 + 30.0 * top++);
  printText(g, s, PFont.helveticaOblique(pdf), 20.0 + 30.0 * top++);
  printText(g, s, PFont.helveticaBoldOblique(pdf), 20.0 + 30.0 * top++);

  printText(g, s, PFont.times(pdf), 20.0 + 30.0 * top++);
  printText(g, s, PFont.timesBold(pdf), 20.0 + 30.0 * top++);
  printText(g, s, PFont.timesItalic(pdf), 20.0 + 30.0 * top++);
  printText(g, s, PFont.timesBoldItalic(pdf), 20.0 + 30.0 * top++);

  printText(g, s, PFont.symbol(pdf), 20.0 + 30.0 * top++);
  printText(g, s, PFont.zapfDingbats(pdf), 20.0 + 30.0 * top++);


  pdf.save('build/ttf-1.pdf');
}
