// ignore_for_file: omit_local_variable_types

part of pdfio;

abstract class PFont extends PObject {
  /// Constructs a [PFont]. This will attempt to map the font from a known
  /// font name to that in Pdf, defaulting to Helvetica if not possible.
  ///
  /// @param name The document name, ie /F1
  /// @param subtype The pdf type, ie /Type1
  /// @param baseFont The font name, ie /Helvetica
  PFont._create(PDocument pdfDocument, {@required this.subtype})
      : assert(subtype != null),
        super(pdfDocument, '/Font') {
    pdfDocument.fonts.add(this);
  }

  factory PFont.courier(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Courier', 0.910, -0.220, const <double>[]);
  }

  factory PFont.courierBold(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Courier-Bold', 0.910, -0.220, const <double>[]);
  }

  factory PFont.courierBoldOblique(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Courier-BoldOblique', 0.910, -0.220, const <double>[]);
  }

  factory PFont.courierOblique(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Courier-Oblique', 0.910, -0.220, const <double>[]);
  }

  factory PFont.helvetica(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Helvetica', 0.931, -0.225, _helveticaWidths);
  }

  factory PFont.helveticaBold(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Helvetica-Bold', 0.962, -0.228, _helveticaBoldWidths);
  }

  factory PFont.helveticaBoldOblique(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Helvetica-BoldOblique', 0.962, -0.228, _helveticaBoldObliqueWidths);
  }

  factory PFont.helveticaOblique(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Helvetica-Oblique', 0.931, -0.225, _helveticaObliqueWidths);
  }

  factory PFont.times(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Times-Roman', 0.898, -0.218, _timesWidths);
  }

  factory PFont.timesBold(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Times-Bold', 0.935, -0.218, _timesBoldWidths);
  }

  factory PFont.timesBoldItalic(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Times-BoldItalic', 0.921, -0.218, _timesBoldItalicWidths);
  }

  factory PFont.timesItalic(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Times-Italic', 0.883, -0.217, _timesItalicWidths);
  }

  factory PFont.symbol(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'Symbol', 1.010, -0.293, _symbolWidths);
  }

  factory PFont.zapfDingbats(PDocument pdfDocument) {
    return PdfType1Font._create(pdfDocument, 'ZapfDingbats', 0.820, -0.143, _zapfDingbatsWidths);
  }

  static const String _cannotDecodeMessage = '''---------------------------------------------
Cannot decode the string to Latin1.
This font does not support Unicode characters.
If you want to use strings other than Latin strings, use a TrueType (TTF) font instead.
See https://github.com/DavBfr/dart_pdf/wiki/Fonts-Management
---------------------------------------------''';

  /// The df type of the font, usually /Type1
  final String subtype;

  String get name => '/F$idr';

  String get fontName => null;

  double get ascent => null;

  double get descent => null;

  static const double defaultGlyphWidth = 0.600;

  /// @param os OutputStream to send the object to
  @override
  void _prepare() {
    super._prepare();

    params['/Subtype'] = PName(subtype);
    params['/Name'] = PName(name);
    params['/Encoding'] = const PName('/WinAnsiEncoding');
  }

  @Deprecated('Use `glyphMetrics` instead')
  double glyphAdvance(int charCode) => glyphMetrics(charCode).advanceWidth;

  PdfFontMetrics glyphMetrics(int charCode);

  @Deprecated('Use `glyphMetrics` instead')
  PRect glyphBounds(int charCode) => glyphMetrics(charCode).toPRect();

  PdfFontMetrics stringMetrics(String s, {double letterSpacing = 0}) {
    if (s.isEmpty) {
      return PdfFontMetrics.zero;
    }

    try {
      final Uint8List chars = latin1.encode(s);
      final Iterable<PdfFontMetrics> metrics = chars.map(glyphMetrics);
      return PdfFontMetrics.append(metrics, letterSpacing: letterSpacing);
    } catch (_) {
      assert(() {
        print(_cannotDecodeMessage);
        return true;
      }());

      rethrow;
    }
  }

  @Deprecated('Use `stringMetrics` instead')
  PRect stringBounds(String s) => stringMetrics(s).toPRect();

  PPoint stringSize(String s) {
    final PdfFontMetrics metrics = stringMetrics(s);
    return PPoint(metrics.width, metrics.height);
  }

  @override
  String toString() => 'Font($fontName)';

  void putText(PStream stream, String text) {
    try {
      PString(latin1.encode(text)).output(stream);
    } catch (_) {
      assert(() {
        print(_cannotDecodeMessage);
        return true;
      }());

      rethrow;
    }
  }
}
