part of pdfio;

class PObject {
  PObject(this.document, [String type]) {
    idr = document.nextIdr();
    if (type != null) params['/Type'] = PName(type);
    document.objects.add(this);
  }

  final PDict params = PDict();
  int idr;
  final int idg = 0;
  final PDocument document;

  void _write(PStream os) {
    _prepare();
    _writeStart(os);
    _writeContent(os);
    _writeEnd(os);
  }

  /// Prepare the object to be written to the stream
  @mustCallSuper
  void _prepare() {}

  void _writeStart(PStream os) {
    os.putString('$idr $idg obj\n');
  }

  void _writeContent(PStream os) {
    if (params.isNotEmpty) {
      params.output(os);
      os.putString('\n');
    }
  }

  void _writeEnd(PStream os) {
    os.putString('endobj\n');
  }

  PIDR idref() => PIDR(idr, idg);
}

///-------------------------------

class PObjects extends PObject {
  PObjects(PDocument document, this.array) : super(document);

  final PArray array;

  @override
  void _writeContent(PStream os) {
    super._writeContent(os);
    array.output(os);
    os.putBytes(<int>[0x0a]);
  }
}

class POStream extends PObject {
  POStream(PDocument pdfDocument, {String type, this.isBinary = false}) : super(pdfDocument, type);
  final PStream stream = PStream();
  final bool isBinary;
  Uint8List _data;
  @override
  void _prepare() {
    super._prepare();

    if (params.containsKey('/Filter') && _data == null) {
      _data = stream.output();
    } else {
      final Uint8List original = stream.output();
      final Uint8List newData = zlib.encode(original);
      if (newData.lengthInBytes < original.lengthInBytes) {
        params['/Filter'] = const PName('/FlateDecode');
        _data = newData;
      }
    }

    if (_data == null) {
      if (isBinary) {
        final Ascii85Encoder e = Ascii85Encoder();
        _data = e.convert(stream.output());
        params['/Filter'] = const PName('/ASCII85Decode');
      } else {
        _data = stream.output();
      }
    }
    params['/Length'] = PNum(_data.length);
  }

  @override
  void _writeContent(PStream os) {
    super._writeContent(os);
    os.putString('stream\n');
    os.putBytes(_data);
    os.putString('\nendstream\n');
  }
}

class XObject extends POStream {
  XObject(PDocument document, String subtype, {bool isBinary = false}) : super(document, type: '/XObject', isBinary: isBinary) {
    params['/Subtype'] = PName(subtype);
  }
}

class FXObject extends XObject {
  FXObject(PDocument pdfDocument) : super(pdfDocument, '/Form') {
    params['/FormType'] = const PNum(1);
    params['/BBox'] = PArray.fromNum(const <int>[0, 0, 1000, 1000]);
  }

  final Map<String, PFont> fonts = <String, PFont>{};
  final Map<String, XObject> xobjects = <String, XObject>{};

  void setMatrix(Matrix4 t) {
    final Float64List s = t.storage;
    params['/Matrix'] = PArray.fromNum(<double>[s[0], s[1], s[4], s[5], s[12], s[13]]);
  }

  @override
  void _prepare() {
    super._prepare();
    final PDict resources = PDict();
    if (fonts.isNotEmpty) resources['/Font'] = PDict.fromObjectMap(fonts);
    if (xobjects.isNotEmpty) resources['/XObject'] = PDict.fromObjectMap(xobjects);
    if (resources.isNotEmpty) params['/Resources'] = resources;
  }
}
