

// ignore_for_file: omit_local_variable_types

part of pdfio;

class POutput {
  POutput(this.os) {
    os.putString('%PDF-${PDocument.version}\n');
    os.putString('%AlmPazel\n');
  }

  final PStream os;
  List<PXref> xrefs = <PXref>[];
  PObject root;
  PObject info;

  void write(PObject ob) {
    if (ob is PRoot) root = ob;
    if (ob is PInfo) info = ob;
    xrefs.add(PXref(ob.idr, os.offset));
    ob._write(os);
  }

  /// This closes the Stream, writing the xref table
  void close() {
    final int xref = os.offset;
    os.putString('xref\n');
    int firstid = 0; // First id in block
    int lastid = -1; // The last id used
    final List<PXref> block = <PXref>[]; // xrefs in this block
    block.add(PXref(0, 0, gen: 65535));

    for (PXref xref in xrefs) {
      if (firstid == -1) firstid = xref.id;
      if (lastid > -1 && xref.id != (lastid + 1)) {
        writeblock(firstid, block);
        block.clear();
        firstid = -1;
      }
      // now add to block
      block.add(xref);
      lastid = xref.id;
    }

    if (firstid > -1) writeblock(firstid, block);

    os.putString('trailer\n');

    var params = PDict();
    params['/Size'] = PNum(xrefs.length + 1);
    params['/Root'] = root.idref();
    var id = PString(root.document.id, true);
    params['/ID'] = PArray(<PType>[id, id]);
    params['/Info'] = info.idref();
    // end the trailer object
    params.output(os);
    os.putString('\nstartxref\n$xref\n%%EOF\n');
  }

  void writeblock(int firstid, List<PXref> block) {
    os.putString('$firstid ${block.length}\n');
    for (PXref x in block) {
      os.putString(x.ref());
      os.putString('\n');
    }
  }
}
